/*
 * Copyright 1993-2015 NVIDIA Corporation.  All rights reserved.
 *
 * Please refer to the NVIDIA end user license agreement (EULA) associated
 * with this source code for terms and conditions that govern your use of
 * this software. Any use, reproduction, disclosure, or distribution of
 * this software and related documentation outside the terms of the EULA
 * is strictly prohibited.
 *
 */

/*
  Parallel reduction kernels
*/

#include <stdio.h>
#include <iostream>
#include <vector>
#include <cooperative_groups.h>

namespace cg = cooperative_groups;

#include "cuda_check.cuh"

__global__ void
dot(const double *g_idata1, const double *g_idata2, 
    double *g_odata, unsigned long n)
{
  cg::thread_block cta = cg::this_thread_block();
  extern __shared__ double sdata[];
  
  size_t tid = threadIdx.x;
  unsigned long i = blockIdx.x*blockDim.x + threadIdx.x;
  
  sdata[tid] = (i < n) ? (g_idata1[i] * g_idata2[i]) : 0;

  cg::sync(cta);

  for (size_t s=blockDim.x/2; s>0; s>>=1)
    {
      if (tid < s)
	      sdata[tid] += sdata[tid + s];

      cg::sync(cta);
    }
  if (tid == 0) g_odata[blockIdx.x] = sdata[0];
}

__global__ void localInit(double* a, int n)
{
  int i = blockDim.x * blockIdx.x + threadIdx.x;
  if (i < n)
    a[i] = 0.0;
}

double dotCuda(unsigned long n,
               const double *d_idata1, const double *d_idata2,
	             size_t threads, size_t blocks)
{  
  double s;
  std::vector<double> s_block(blocks);
  double *d_temp;
  cudaMalloc(&d_temp, blocks*sizeof(double));

  size_t smemSize = (threads <= 32)
    ? 2 * threads * sizeof(double)
    : threads * sizeof(double);
  
  dot<<< blocks, threads, smemSize >>>(d_idata1, d_idata2, d_temp, n);
  CUDA_CHECK_KERNEL();
  
  CUDA_CHECK_KERNEL();
  CUDA_CHECK_OP(cudaMemcpy(s_block.data(),
  			   d_temp,
  			   blocks * sizeof(double),
  			   cudaMemcpyDeviceToHost));

  s = 0.0;
  for (size_t i=0; i<blocks; i++)
    s += s_block[i];

  cudaFree(d_temp);
  return s;
}
