#include <iostream>
#include <iomanip>

#include <vector>
#include "Matrix.hxx"

#include "Vector.cuh"
#include "Matrix.cuh"
#include "dot.h"
#include "cuda_check.cuh"
#include "arguments.hxx"
#include "timer.hxx"
#include "info.hxx"

const char * version = VERSION;

__global__ void dprintf(const double *v)
{
  printf(" %10g", *v);
}

void dprint(const CudaVector & v)
{
  size_t i, n=v.n();
  const double * x = v.data();
  for (i=0; i<n; i++, x++) {
    printf("%5ld: ", i);
    dprintf<<<1,1>>>(x);
    cudaDeviceSynchronize();
    printf("\n");
  }
  printf("\n");
}

void dprint(const CudaMatrix & M)
{
  size_t i, j, n=M.n(), m = M.m();
  const double * x = M.data();
 
  for (i=0; i<n; i++) {
    printf("%5ld: ", i);
    for (j=0; j<m; j++, x++) {
      dprintf<<<1,1>>>(x);
      cudaDeviceSynchronize();
    }
    printf("\n");
  }
  printf("\n");
}

std::ostream & operator<<(std::ostream &f, const std::vector<double> &x)
{
  size_t i, n=x.size();
  for (i=0; i<n; i++)
    f << std::setw(5) << i << ": " << std::setw(10) << x[i] << std::endl;
  return f;
}

std::ostream & operator<<(std::ostream &f, const Matrix &M)
{
  size_t i, j, n=M.n(), m=M.m();
  const double *c = M.data();

  for (i=0; i<n; i++) {
    f << std::setw(5) << i << ": ";
    for (j=0; j<m; j++)
      f << std::setw(10) << c[i*m + j];
    f << std::endl;
  }
  return f;
}

__global__ void initCuda(double* a, double value, int n)
{
  int i = blockDim.x * blockIdx.x + threadIdx.x;
  if (i < n) {
    a[i] = value * i;
  }
}


void init(CudaMatrix &M, double x1, 
          CudaVector &v, double x2,
          CudaVector &w, double x3)
{
  Timer t;
  t.start();

  size_t n = M.n(), m = M.m(), nm = n*m;
  double y1 = x1/(nm-1);
  double y2 = x2/(m-1);
  double y3 = x3/(n-1);

  initCuda<<<(nm+128)/128, 128>>>(M.data(), y1, nm);
  CUDA_CHECK_KERNEL();
  initCuda<<<(m+128)/128, 128>>>(v.data(), y2, m);
  CUDA_CHECK_KERNEL();
  initCuda<<<(n+128)/128, 128>>>(w.data(), y3, n);
  CUDA_CHECK_KERNEL();

  t.stop();
  std::cout << "Initialisations : " << t.elapsed() << " s" << std::endl;
}

__global__ void matvecCuda(double *w, const double *A, const double *v, int n, int m)
{
  int i = blockDim.x * blockIdx.x + threadIdx.x, j;
  double s = 0.0;
  if (i < n) {
    for (j=0; j<m; j++)
      s += A[i*m + j] * v[j];
    w[i] = s;
  }
}

void matvec(CudaVector &w, const CudaMatrix &A, const CudaVector &v)
{
  Timer t;
  t.start();

  size_t n = A.n(), m = A.m() ;
  const double *x = v.data();
  double * y = w.data();
  const double *M = A.data();

  matvecCuda<<<(n + 64)/64, 64>>>(y, M, x, n, m);
  CUDA_CHECK_KERNEL();

  t.stop();
  std::cout << "Produit matrice - vecteur : " << t.elapsed() << " s"  << std::endl;

}

double dot(const CudaVector & d_x,
           const CudaVector & d_y)
{
  Timer t;
  t.start();
  size_t n = d_x.n();
  size_t blockSize = 32;
  size_t gridSize = ceil((n+blockSize)/blockSize);

  double result =  dotCuda(n, d_x.data(), d_y.data(),
 	                         blockSize, gridSize);

  t.stop();
  std::cout << "Produit scalaire vecteurs : " << t.elapsed() << " s"  << std::endl;
  return result;
}

int main(int argc, char** argv)
{
  Arguments Args(argc, argv);
  size_t n = Args.Get("n", 10000L);
  size_t m = Args.Get("m", n);

  std::cout << "version " << version << "\n\nn,m = " << n << " " << m << "\n\n";

  size_t m_max = doubleAvail();
  if (2*m + n + n*m > 0.8*m_max) {
    std::cerr << "insufficient memoy available (" << m_max << "doubles)" << std::endl;
    return -1;
  }

  CudaVector d_x(m);
  CudaVector d_temp(m);
  CudaVector d_y(n);
  CudaMatrix d_A(n, m);

  init(d_A, 2.0, d_x, 1.0, d_y, 1.5);

  matvec(d_temp, d_A, d_x);

  double s = dot(d_y, d_temp);
  std::cout << "\n\tyAx " << s << std::endl;

  if (n <= 10 && m <= 10) {
    Timer t_get;
    t_get.start();
    std::vector<double> h_y(n), h_x(m);
    Matrix h_A(n, m);
    cudaMemcpy(h_x.data(), d_x.data(), d_x.bytes(), cudaMemcpyDeviceToHost);
    cudaMemcpy(h_y.data(), d_y.data(), d_y.bytes(), cudaMemcpyDeviceToHost);
    cudaMemcpy(h_A.data(), d_A.data(), d_A.bytes(), cudaMemcpyDeviceToHost);
    t_get.stop();
    std::cout << "Copie (GPU -> CPU) :              " << t_get.elapsed() << " s"  << std::endl;
    std::cout << "x\n" << h_x << std::endl;
    std::cout << "A\n" << h_A << std::endl;
    std::cout << "y\n" << h_y << std::endl;
  }
  std::cout << "\n";
  return 0;
}
